# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Defines namespace utilities used by the package.
"""

__all__ = []


from .instances import *
from .slots import *
from .typed_slots import *
from .namespaces import *
from .configs import *
from .packaging import *


modules = dict(
    configs=configs,
    instances=instances,
    namespaces=namespaces,
    packaging=packaging,
    slots=slots,
    typed_slots=typed_slots,
)


module_api = dict(
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)
